/********************************************************************************************
*********************************************************************************************
	"Stored Procedures Solutions", script version 1.0
	by Jen McCown, Jen@MidnightDBA.com, http://www.MidnightDBA.com/Jen 
	Session downloads and videos: http://www.midnightdba.com/Jen/articles/

		MidnightSQL Consulting, LLC: 
		When your database needs a hero.

********************************************************************************************
********************************************************************************************/


/**************************** SETUP ****************************/
USE master;
GO
IF EXISTS ( SELECT  *
            FROM    sys.databases 
            WHERE   name = 'Demo' ) 
    DROP DATABASE DEMO;
GO

CREATE DATABASE Demo;
go

USE Demo;
GO

CREATE SCHEMA Audit;
GO


/**************************** INTRO ****************************/
/*
	Today's topic: Developing stored procedure solutions.

	What's an SP? Reusable code, stored in a database.
*/



/**************************** SCENARIO ****************************/
/*
	Our scenario: We want to track database files sizes over time.

	There's more than one solution. What's best? (We're creating a
	recipe, a means to an end.)
*/


/**************************** Step 1: QUESTIONS ****************************/

-- FIRST, ESSENTIAL HOW TO - Where do we even get file size information?
SELECT  *
FROM    sys.master_files;

/*
QnA: Example questions and answers:
	 Do we have an existing solution?

	 Do we need this?

	 Why do we need this? What will it be used for?

	 What else could it be used for?

	 Who will use it? Would anyone else find it useful (if appropriate)?

	 How often might it be used?

	 How big could it get? Scope, size, user base, frequency, number of servers, etc.
		
	 How might it be used?

	 Security concerns? Is this secure data? Who should/n't have access?

	 Where does the solution belong? Examples: Ad hoc SQL, Excel, SP, Powershell, SSRS, Program, elsewhere.

---

Example questions and answers:
	 Do we have an existing solution?
		Nope.

	 Do we need this?
		Yep.

	 Why do we need this? What will it be used for?
		We want to know how much space the DB files are taking up now, and
		what rate of growth they show. Will be useful for disk capacity planning.

	 What else could it be used for?
		We'll see. Maybe spot checking sizes (not just logging over time)...

	 Who will use it? Would anyone else find it useful (if appropriate)?
		DBAs mostly. SAN guys might like to have access to that info. Maybe
		some managers too, if they don't have to do anything technical.

	 How often might it be used?
		Gather ~weekly or monthly. Ad hoc as needed, but not more than a handful of
		times per day at most.

	 How big could it get? Scope, size, user base, frequency, number of servers, etc.
		Limited scope and user base. Better include a cleanup mechanism for size.
		Frequency ~wk or mo. Servers, hmm...
		
	 How might it be used?

	 Security concerns?

	 Where does the solution belong? Examples: Ad hoc SQL, Excel, SP, Powershell, SSRS, Program, elsewhere.
		"Tracking" means keeping information over time so we can compare and use it. To us, that means a table.
		Anything that happens repeatedly over time MUST be automated. Simplify your life. That's a job.


**************************** TIME OUT: The Case for SPs ****************************

	Procedural:	  SPs do stuff (select, modify, insert, generate, set, report, alert, etc.) 
				  As opposed to views, functions.

	Stored:		  Stored in the database. Harder to lose it.

	Contractural: Consistent inputs, consistent outputs. What happens in the middle is obfuscated.
	
	Configurable: Pass in values to get different behavior (say, just one or two DBs insted of all).
				  That simplifies the situation where we might want to track one database at a
				  different interval; instead of copying code to a separate job, just call the same
				  SP with different inputs.
	
	Maintainable: A stored procedure is centrally stored, secure, backed up. It eliminates the need of
				  duplicate code stored all over the place (jobs, dev computers, etc.)

	Secure:		  Great for limiting how and what people can do.

	Small footprint: Tiny network calls.
		
*/


/**************************** Step 2: SIMPLEST CASE/POC ****************************/
/*
Simplest case: table, SP, job.
	
*/

EXEC sp_help "sys.master_files";

-- Copied the specs from sys.master_files:
CREATE TABLE AUDIT.DBFileSize
    (
      [database_id] INT NOT NULL ,
      [file_id] INT NOT NULL ,
      [file_guid] UNIQUEIDENTIFIER NULL ,
      [type] TINYINT NOT NULL ,
      [type_desc] NVARCHAR(120) NULL ,
      [data_space_id] INT NOT NULL ,
      [name] SYSNAME NOT NULL ,
      [physical_name] NVARCHAR(520) NOT NULL ,
      [state] TINYINT NULL ,
      [state_desc] NVARCHAR(120) NULL ,
      [size] INT NOT NULL ,
      [max_size] INT NOT NULL ,
      [growth] INT NOT NULL ,
      [is_media_read_only] BIT NOT NULL ,
      [is_read_only] BIT NOT NULL ,
      [is_sparse] BIT NOT NULL ,
      [is_percent_growth] BIT NOT NULL ,
      [is_name_reserved] BIT NOT NULL ,
      [create_lsn] NUMERIC(25,0) NULL ,
      [drop_lsn] NUMERIC(25,0) NULL ,
      [read_only_lsn] NUMERIC(25,0) NULL ,
      [read_write_lsn] NUMERIC(25,0) NULL ,
      [differential_base_lsn] NUMERIC(25,0) NULL ,
      [differential_base_guid] UNIQUEIDENTIFIER NULL ,
      [differential_base_time] DATETIME NULL ,
      [redo_start_lsn] NUMERIC(25,0) NULL ,
      [redo_start_fork_guid] UNIQUEIDENTIFIER NULL ,
      [redo_target_lsn] NUMERIC(25,0) NULL ,
      [redo_target_fork_guid] UNIQUEIDENTIFIER NULL ,
      [backup_lsn] NUMERIC(25,0) NULL
    );
GO


/*
	--
	PRO TIP: Get used to ending your statements with a semicolon (;). 
			 Useful now, and bets are heavy that it'll eventually be mandatory.
	--
*/




CREATE PROCEDURE AUDIT.GetDBFileSize
/*
	Purpose: Log the current file sizes to the AUDIT.DBFileSize table, 
			 for use in disk space projections.
*/ 
AS
    BEGIN
        INSERT  INTO AUDIT.DBFileSize
                SELECT  *
                FROM    sys.master_files;

    END
GO

EXEC AUDIT.GetDBFileSize;

SELECT * FROM AUDIT.DBFileSize;

--*-- Check out SQL Agent to see our job and weekly schedule: "DBA - Audit DB File Size" --*--





/**************************** Step 3: EVALUATE ****************************/
/**************************** Step 3a: Evaluate the table */

DROP TABLE AUDIT.DBFileSize
GO

/*
 Evaluate the table based on need, and on Most Excellent Practices (MEP):
	 Third normal form (3NF)	~=	Primary key, 1 column contains 1 value, 1 row = 1 thing, no duplicate data between rows.
	 Solid naming scheme (consistency, meaning, usability)
	 Data types ("domain")
	 NULL-ability
	 Version compatibility
	 Measurement units (the "size" column here is measured in 8-KB pages, not MB)
	 Too much data? Not enough?

 Discuss: Is it ever okay to break these rules?
 
 (Notice we're not looking at performance, YET.)
*/


CREATE TABLE AUDIT.DBFileSize
    (
      ID INT IDENTITY(1, 1)
             PRIMARY KEY CLUSTERED ,			-- Added a PK.
      [database_name] SYSNAME NOT NULL ,		-- ADDED DB name; names and IDs could change over time.
      [database_id] INT NOT NULL ,
      [file_id] INT NOT NULL ,
												-- REMOVED [file_guid]; don't care. (For one thing, master/model/tempdb files don't have a file_guid.)
      [type_desc] NVARCHAR(120) NULL ,			-- Kept [type_desc] instead of [type].
      [data_space_id] INT NOT NULL ,			-- Refers to filegroup. Do we care? Maybe?
      [name] SYSNAME NOT NULL ,
      [physical_name] NVARCHAR(520) NOT NULL ,
      [state_desc] NVARCHAR(120) NULL ,			-- Kept [state_desc] instead of [state].
      [size] INT NOT NULL ,
      [sizeMB] AS ( size * 8 / 1024 ) PERSISTED ,	-- Added a computed, persisted column for file size in MB. 

	  -- FILE GROWTH INFO:
      [max_size] INT NOT NULL ,
      [growth] INT NOT NULL ,
      [is_media_read_only] BIT NOT NULL ,
      [is_read_only] BIT NOT NULL ,
      [is_sparse] BIT NOT NULL ,
      [is_percent_growth] BIT NOT NULL ,

	  sample_date DATETIME2 DEFAULT (GETDATE()) NULL	-- Note that DATETIME2 was introduced in SQL 2008; won't work on 2005 servers.
      
	  -- Dropped [is_name_reserved] and all the LSN data.
    );
GO




/**************************** Step 3b: Evaluate the SP */
/*
 Evaluate the SP based on need, and on Most Excellent Practices (MEP):
	 NEVER EVER use SELECT * in a query, especially not in an SP.
	 ALWAYS use a column list with INSERT statements. Always.
*/


ALTER PROCEDURE AUDIT.GetDBFileSize
/*
	Purpose: Log the current file sizes to the AUDIT.DBFileSize table, 
			 for use in disk space projections.
*/ 
AS
    BEGIN
        INSERT  INTO AUDIT.DBFileSize
                ( database_name ,
                  database_id ,
                  file_id ,
                  type_desc ,
                  data_space_id ,
                  name ,
                  physical_name ,
                  state_desc ,
                  size ,
                  max_size ,
                  growth ,
                  is_media_read_only ,
                  is_read_only ,
                  is_sparse ,
                  is_percent_growth
                )
                SELECT  d.name AS database_name ,
                        m.database_id ,
                        m.file_id ,
                        m.type_desc ,
                        m.data_space_id ,
                        m.name ,
                        m.physical_name ,
                        m.state_desc ,
                        m.size ,
                        m.max_size ,
                        m.growth ,
                        m.is_media_read_only ,
                        m.is_read_only ,
                        m.is_sparse ,
                        m.is_percent_growth
                FROM    sys.master_files AS m
                        INNER JOIN sys.databases d ON m.database_id = d.database_id;

    END
GO

EXEC AUDIT.GetDBFileSize;

SELECT * FROM AUDIT.DBFileSize;

GO



/**************************** Step 4: EXPAND ****************************/

/*
 Expansions: What about ad hoc? Or, single DB? And what about cleanup?
 Decided: Single SP, instead of multiple for slightly different uses.

 1. Ad hoc use: add a parameter to deal with ad hoc use. 
	Good idea: default behavior should be impermanent (here, not logging data).

		--
		PRO TIP: Limit the input of your SPs, by type (note we used a bit, not 
		an INT or a string type) and/or by programming.

		PRO TIP: Document your SPs! Keep a header, note what parameters are for, etc.
		--
*/


ALTER PROCEDURE AUDIT.GetDBFileSize 
	@log_to_table BIT = 0

/*
	Purpose: Log the current file sizes to the AUDIT.DBFileSize table, 
			 for use in disk space projections.

	Parameters:
		@log_to_table		If 0, just returns file size data (for ad hoc use).
							If 1, logs data to "AUDIT.DBFileSize".
		
	Example execution:
		EXEC AUDIT.GetDBFileSize;						-- Ad hoc (non logged), all databases.
		EXEC AUDIT.GetDBFileSize @log_to_table = 1;		-- Logged, all databases.
		
*/
AS 
    BEGIN
        IF @log_to_table = 0 
            SELECT  d.name AS database_name ,
                    m.database_id ,
                    m.file_id ,
                    m.type_desc ,
                    m.data_space_id ,
                    m.name ,
                    m.physical_name ,
                    m.state_desc ,
                    m.size ,
                    m.max_size ,
                    m.growth ,
                    m.is_media_read_only ,
                    m.is_read_only ,
                    m.is_sparse ,
                    m.is_percent_growth
            FROM    sys.master_files AS m
                    INNER JOIN sys.databases d ON m.database_id = d.database_id;
        ELSE IF @log_to_table = 1
            INSERT  INTO AUDIT.DBFileSize
                    ( database_name ,
                      database_id ,
                      file_id ,
                      type_desc ,
                      data_space_id ,
                      name ,
                      physical_name ,
                      state_desc ,
                      size ,
                      max_size ,
                      growth ,
                      is_media_read_only ,
                      is_read_only ,
                      is_sparse ,
                      is_percent_growth
                    )
                    SELECT  d.name AS database_name ,
                            m.database_id ,
                            m.file_id ,
                            m.type_desc ,
                            m.data_space_id ,
                            m.name ,
                            m.physical_name ,
                            m.state_desc ,
                            m.size ,
                            m.max_size ,
                            m.growth ,
                            m.is_media_read_only ,
                            m.is_read_only ,
                            m.is_sparse ,
                            m.is_percent_growth
                    FROM    sys.master_files AS m
                            INNER JOIN sys.databases d ON m.database_id = d.database_id;

	
    END
GO

/*
 2. Option for single DB. Add a parameter.

		--
		PRO TIP: Parameter / variable lengths can be tricky. Think expansively, and long term.
		--

*/

ALTER PROCEDURE AUDIT.GetDBFileSize 
	@log_to_table BIT = 0 ,
	@database_filter NVARCHAR(256) = '%'
/*
	Purpose: Log the current file sizes to the AUDIT.DBFileSize table, 
			 for use in disk space projections.

	Parameters:
		@log_to_table		If 0, just returns file size data (for ad hoc use).
							If 1, logs data to "AUDIT.DBFileSize".
		
		@database_filter	Filters the database data to retrieve.  Defaults to all ('%'). 

		
	Example execution:
		EXEC AUDIT.GetDBFileSize;						-- Ad hoc (non logged), all databases.
		EXEC AUDIT.GetDBFileSize @log_to_table = 1;		-- Logged, all databases.
		EXEC AUDIT.GetDBFileSize @log_to_table = 0, 
								 @database_filter = 'Adventure%';	-- Nonlogged, all "Adventure*" databases.
*/
AS 
    BEGIN
        IF @log_to_table = 0 
            SELECT  d.name AS database_name ,
                    m.database_id ,
                    m.file_id ,
                    m.type_desc ,
                    m.data_space_id ,
                    m.name ,
                    m.physical_name ,
                    m.state_desc ,
                    m.size ,
                    m.max_size ,
                    m.growth ,
                    m.is_media_read_only ,
                    m.is_read_only ,
                    m.is_sparse ,
                    m.is_percent_growth
            FROM    sys.master_files AS m
                    INNER JOIN sys.databases d ON m.database_id = d.database_id
			WHERE d.NAME LIKE @database_filter; 
        ELSE 
            INSERT  INTO AUDIT.DBFileSize
                    ( database_name ,
                      database_id ,
                      file_id ,
                      type_desc ,
                      data_space_id ,
                      name ,
                      physical_name ,
                      state_desc ,
                      size ,
                      max_size ,
                      growth ,
                      is_media_read_only ,
                      is_read_only ,
                      is_sparse ,
                      is_percent_growth
                    )
                    SELECT  d.name AS database_name ,
                            m.database_id ,
                            m.file_id ,
                            m.type_desc ,
                            m.data_space_id ,
                            m.name ,
                            m.physical_name ,
                            m.state_desc ,
                            m.size ,
                            m.max_size ,
                            m.growth ,
                            m.is_media_read_only ,
                            m.is_read_only ,
                            m.is_sparse ,
                            m.is_percent_growth
                    FROM    sys.master_files AS m
                            INNER JOIN sys.databases d ON m.database_id = d.database_id
					WHERE d.NAME LIKE @database_filter;

    END
GO

/**************************** Exploring options: Table Valued Parameters */
-- You do have the option to read in a comma delimited list of values, parse and use them; 
-- or, a table valued parameter (SQL 2008 +, requires a custom type).


CREATE TYPE dbo.DBlist AS TABLE 
( NAME SYSNAME ); 
GO

CREATE PROCEDURE AUDIT.GetDBFileSize_TVPexample
    @log_to_table BIT = 0 ,
    @database_filter DBlist READONLY
/*
	Purpose: Log the current file sizes to the AUDIT.DBFileSize table, 
			 for use in disk space projections.

	Parameters:
		@log_to_table		If 0, just returns file size data (for ad hoc use).
							If 1, logs data to "AUDIT.DBFileSize".
		
		@database_filter	Defaults to all ('%'). 

	Example execution:
		EXEC AUDIT.GetDBFileSize_TVPexample;						-- Ad hoc (non logged), all databases.
		EXEC AUDIT.GetDBFileSize_TVPexample @log_to_table = 1;		-- Logged, all databases.
		EXEC AUDIT.GetDBFileSize_TVPexample @log_to_table = 1, 
								 @database_filter = 'Adventure%';	-- Logged, all "Adventure*" databases.
*/
AS 
    BEGIN
        IF @log_to_table = 0 
            SELECT  d.name AS database_name ,
                    m.database_id ,
                    m.file_id ,
                    m.type_desc ,
                    m.data_space_id ,
                    m.name ,
                    m.physical_name ,
                    m.state_desc ,
                    m.size ,
                    m.max_size ,
                    m.growth ,
                    m.is_media_read_only ,
                    m.is_read_only ,
                    m.is_sparse ,
                    m.is_percent_growth
            FROM    sys.master_files AS m
                    INNER JOIN sys.databases d ON m.database_id = d.database_id
                    INNER JOIN @database_filter db ON db.NAME = d.name;
        ELSE  IF @log_to_table = 1
            INSERT  INTO AUDIT.DBFileSize
                    ( database_name ,
                      database_id ,
                      file_id ,
                      type_desc ,
                      data_space_id ,
                      name ,
                      physical_name ,
                      state_desc ,
                      size ,
                      max_size ,
                      growth ,
                      is_media_read_only ,
                      is_read_only ,
                      is_sparse ,
                      is_percent_growth
                    )
                    SELECT  d.name AS database_name ,
                            m.database_id ,
                            m.file_id ,
                            m.type_desc ,
                            m.data_space_id ,
                            m.name ,
                            m.physical_name ,
                            m.state_desc ,
                            m.size ,
                            m.max_size ,
                            m.growth ,
                            m.is_media_read_only ,
                            m.is_read_only ,
                            m.is_sparse ,
                            m.is_percent_growth
                    FROM    sys.master_files AS m
                            INNER JOIN sys.databases d ON m.database_id = d.database_id
                            INNER JOIN @database_filter db ON db.NAME = d.name;

    END
GO

---- Run with values:
DECLARE @dblist DBlist;

INSERT  INTO @dblist
        ( name )
VALUES  ( 'master' ),
        ( 'msdb' ),
        ( 'model' );

EXEC AUDIT.GetDBFileSize_TVPexample @log_to_table = 0, @database_filter = @dblist;

GO
---- Run without values: 
DECLARE @dblist DBlist;

EXEC AUDIT.GetDBFileSize_TVPexample @log_to_table = 0, @database_filter = @dblist;
GO



/*
 3. Cleanup! Add a parameter for retention length.

	-- 
	PRO TIP: Test potentially confusing code a few different ways.
	--
*/

ALTER PROCEDURE AUDIT.GetDBFileSize 
	@log_to_table BIT = 0 ,
	@database_filter NVARCHAR(2000) = '%', 
	@retention_days SMALLINT = 730
/*
	Purpose: Log the current file sizes to the AUDIT.DBFileSize table, 
			 for use in disk space projections.

	Parameters:
		@log_to_table		If 0, just returns file size data (for ad hoc use).
							If 1, logs data to "AUDIT.DBFileSize".
		
		@database_filter	Filters the database data to retrieve.  Defaults to all ('%'). 

		@retention_days		Days to retain data in the AUDIT.DBFileSize table. Default is 2 years (730 days).

	Example execution:
		EXEC AUDIT.GetDBFileSize;						-- Ad hoc (non logged), all databases.
		EXEC AUDIT.GetDBFileSize @log_to_table = 1;		-- Logged, all databases.
		EXEC AUDIT.GetDBFileSize @log_to_table = 1, 
								 @database_filter = 'Adventure%';	-- Logged, all "Adventure*" databases.
*/
AS 
    BEGIN
        IF @log_to_table = 0 
            SELECT  d.name AS database_name ,
                    m.database_id ,
                    m.file_id ,
                    m.type_desc ,
                    m.data_space_id ,
                    m.name ,
                    m.physical_name ,
                    m.state_desc ,
                    m.size ,
                    m.max_size ,
                    m.growth ,
                    m.is_media_read_only ,
                    m.is_read_only ,
                    m.is_sparse ,
                    m.is_percent_growth
            FROM    sys.master_files AS m
                    INNER JOIN sys.databases d ON m.database_id = d.database_id
			WHERE d.NAME LIKE @database_filter;
        ELSE 
            INSERT  INTO AUDIT.DBFileSize
                    ( database_name ,
                      database_id ,
                      file_id ,
                      type_desc ,
                      data_space_id ,
                      name ,
                      physical_name ,
                      state_desc ,
                      size ,
                      max_size ,
                      growth ,
                      is_media_read_only ,
                      is_read_only ,
                      is_sparse ,
                      is_percent_growth
                    )
                    SELECT  d.name AS database_name ,
                            m.database_id ,
                            m.file_id ,
                            m.type_desc ,
                            m.data_space_id ,
                            m.name ,
                            m.physical_name ,
                            m.state_desc ,
                            m.size ,
                            m.max_size ,
                            m.growth ,
                            m.is_media_read_only ,
                            m.is_read_only ,
                            m.is_sparse ,
                            m.is_percent_growth
                    FROM    sys.master_files AS m
                            INNER JOIN sys.databases d ON m.database_id = d.database_id
					WHERE d.NAME LIKE @database_filter;

        DELETE  FROM AUDIT.DBFileSize
        WHERE   sample_date < DATEADD(DAY, -1 * @retention_days, GETDATE());
		
    END
GO
/*
 Figure out and test the code:
	 I want to delete WHERE the sample_date is farther back from (today minus retention days)
	  =>
	 Delete WHERE sample_date < (today minus retention days)
	  =>
	 WHERE sample_date < GetDate() - @retention_days; -- this is legal, but less explicit visually. Arguable.
	  =>
	 WHERE sample_date < DATEADD(DAY, -1 * @retention_days, GetDate());
*/	
	---- TESTING ---- 
	DECLARE @example_sample_date DATETIME;
	DECLARE @retention_days SMALLINT;
	SET @retention_days = (365 * 2);
	
	-- Today, and the statement should come back false --
	SET @example_sample_date = GETDATE();

	IF (@example_sample_date < DATEADD(DAY, -1 * @retention_days, GetDate())) -- This should turn out untrue.
		SELECT 'FAIL. Example date (today) is less than the earliest retention date? Nope, logic is flawed.';
	ELSE
		SELECT 'PASS. Today is not earlier than our retention period.'  


	-- One year ago --
	SET @example_sample_date = GETDATE() - 365;

	IF (@example_sample_date < DATEADD(DAY, -1 * @retention_days, GetDate()))
		SELECT 'FAIL. Example date (1 year go) is less than the earliest retention date? Nope, logic is flawed.';
	ELSE
		SELECT 'PASS. 1 year ago is not earlier than our retention period.';


	-- Three years ago --
	SET @example_sample_date = GETDATE() - (365 * 3);

	IF (@example_sample_date < DATEADD(DAY, -1 * @retention_days, GetDate()))
		SELECT 'PASS. Three years ago is earlier than two years ago.'
	ELSE
		SELECT 'FAIL. Flawed logic.'  

-- Alternate testing:
    TRUNCATE TABLE AUDIT.DBFileSize;

	EXEC AUDIT.GetDBFileSize @log_to_table = 1;

    UPDATE  AUDIT.DBFileSize
    SET     sample_date = GETDATE() - 5;

    EXEC AUDIT.GetDBFileSize @log_to_table = 1;
    SELECT  COUNT(*)
    FROM    AUDIT.DBFileSize;

    UPDATE  AUDIT.DBFileSize
    SET     sample_date = GETDATE() - 8
    WHERE   sample_date > GETDATE() - 4;

    SELECT DISTINCT
            DATEDIFF(day, sample_date, GETDATE()) AS Days_Ago ,
            sample_date
    FROM    audit.dbfilesize;

EXEC AUDIT.GetDBFileSize @log_to_table = 0, @retention_days = 10; -- Should delete nothing
SELECT @@rowcount AS Rows_Deleted; 


EXEC AUDIT.GetDBFileSize @log_to_table = 0, @retention_days = 6; -- Should delete 47 rows
SELECT @@rowcount AS Rows_Deleted;


EXEC AUDIT.GetDBFileSize @log_to_table = 0, @retention_days = 1; -- Should delete 47 more rows
SELECT @@rowcount AS Rows_Deleted;



/**************************** Step 5: EVALUATE FOR PERFORMANCE ****************************/
/*
 Performance is a HUGE topic, so go read many books. 
 BUT, there are some Most Excellent Practices:
	
	Everything we've said: 3NF for tables, consistency, etc.
	Avoid cursors
	Avoid query hints
	Test your code with realistic data; examine query plans
	Index appropriately
	Watch for parameter sniffing

	(Breaking rules)
*/


/**************************** Step N: ...AND MORE AND MORE ****************************/
/*
Additional topics (maybe part 2):
	 Error handling
	 Expansion: File space usage 
	 Code review
	 More testing
	 Dynamic T-SQL

	 Multi server coding! - How? You'll need to come to our Enterprise Management session!

*/




/********************************************************************************************
*********************************************************************************************
This session, recording and materials are licensed under a Creative Commons Attribution-
Noncommercial-No Derivative Works 3.0 United States License.
********************************************************************************************
********************************************************************************************/